#ifdef _MSC_VER
#pragma warning(disable : 4786)
#endif // _MSC_VER

///////////////////////////////////////////////////////////////////////////////
// include

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <direct.h>
#include <io.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <windows.h>

#include <vector>
#include <map>
#include <string>

///////////////////////////////////////////////////////////////////////////////
// basic type

typedef BYTE uint8_t;
typedef WORD uint16_t;
typedef DWORD uint32_t;

///////////////////////////////////////////////////////////////////////////////
// basic function

#define unlink	_unlink

///////////////////////////////////////////////////////////////////////////////
// data type

enum ExtractResultType {
	EXERR_NOERROR    = 0,
	EXERR_NOTFOUND   = -1,
	EXERR_NOTCREATE  = -2,
	EXERR_IOERROR    = -3,
	EXERR_READERROR  = -4,
	EXERR_WRITEERROR = -5,
	EXERR_BADFORMAT  = -6
};

struct RangeInfo {
	uint32_t ofs, len;
};

struct DirInfo {
	std::string name;
	RangeInfo range;
};

typedef std::map<std::string, int> DirMap;

typedef std::map<std::string, std::string> ConvertMap;

///////////////////////////////////////////////////////////////////////////////
// registry function

HKEY openRegistryKey(std::string section, bool writable) {
	HKEY parentKey = NULL,
		childKey = HKEY_CURRENT_USER;
	DWORD options = KEY_READ | (writable? KEY_WRITE: 0);
	const char* subKeyBegin = section.c_str();
	while (true) {
		if ((parentKey != NULL) &&
			(parentKey != HKEY_CURRENT_USER))
		{
			RegCloseKey(parentKey);
		}
		if (childKey == NULL) {
			break;
		}
		if (*subKeyBegin == '\0') {
			break;
		}
		parentKey = childKey;
		childKey = NULL;
		const char* subKeyEnd = subKeyBegin;
		while (
			(*subKeyEnd != '\0') &&
			(*subKeyEnd != '\\'))
		{
			subKeyEnd = CharNext(subKeyEnd);
		}
		if (subKeyEnd == subKeyBegin) {
			break;
		}
		std::string subKey(subKeyBegin, subKeyEnd);
		if (*subKeyEnd == '\\') {
			subKeyEnd = CharNext(subKeyEnd);
		}
		subKeyBegin = subKeyEnd;
		if (!writable) {
			if (RegOpenKeyEx(
					parentKey, subKey.c_str(), 0,
					options,
					&childKey) != ERROR_SUCCESS)
			{
				childKey = NULL;
			}
		} else {
			DWORD dwDP;
			if (RegCreateKeyEx(
					parentKey, subKey.c_str(), 0, REG_NONE,
					REG_OPTION_NON_VOLATILE,
					options, NULL,
					&childKey, &dwDP) != ERROR_SUCCESS)
			{
				childKey = NULL;
			}
		}
	}
	return childKey;
}

std::string getRegistryString(
	const std::string section,
	const std::string entry,
	const std::string def)
{
	std::string value = def;
	HKEY sectionKey = openRegistryKey(section, false);
	if (sectionKey != NULL) {
		DWORD type, count;
		count = sizeof(DWORD);
		LONG result = RegQueryValueEx(
			sectionKey, entry.c_str(), NULL,
			&type, NULL, &count);
		if ((result == ERROR_SUCCESS) &&
			((type == REG_SZ) || (type == REG_BINARY)))
		{
			char* tmp = new TCHAR[(size_t)count+1];
			if (tmp != NULL) {
				tmp[count] = '\0';
				result = RegQueryValueEx(
					sectionKey, entry.c_str(), NULL,
					&type, (BYTE*)tmp, &count);
				if (result == ERROR_SUCCESS) {
					value = tmp;
				}
				delete [] tmp;
			}
		}
		RegCloseKey(sectionKey);
	}
	return value;
}

///////////////////////////////////////////////////////////////////////////////
// file function

bool existFile(const char* fileName) {
	bool exist = false;
	FILE* fpt = fopen(fileName, "rb");
	if (fpt != NULL) {
		exist = true;
		fclose(fpt);
	}
	return exist;
}

void readBytes(FILE* fpt, void* buf, size_t len) {
	if (fread(buf, len, 1, fpt) != 1) {
		throw EXERR_READERROR;
	}
}

uint8_t readUINT8(FILE* fpt) {
	uint8_t data;
	readBytes(fpt, &data, sizeof(data));
	return data;
}

uint16_t readUINT16(FILE* fpt) {
	uint16_t data;
	readBytes(fpt, &data, sizeof(data));
	return data;
}

uint32_t readUINT32(FILE* fpt) {
	uint32_t data;
	readBytes(fpt, &data, sizeof(data));
	return data;
}

void writeBytes(FILE* fpt, const void* buf, size_t len) {
	if (fwrite(buf, len, 1, fpt) != 1) {
		throw EXERR_READERROR;
	}
}

void writeUINT8(FILE* fpt, uint8_t data) {
	writeBytes(fpt, &data, sizeof(data));
}

void writeUINT16(FILE* fpt, uint16_t data) {
	writeBytes(fpt, &data, sizeof(data));
}

void writeUINT32(FILE* fpt, uint32_t data) {
	writeBytes(fpt, &data, sizeof(data));
}

void convertPackFile(
	const char* dstFileName,
	const char* srcFileName,
	const ConvertMap& cmap)
{
	FILE* sfpt = NULL;
	FILE* dfpt = NULL;
	static char buf[4096];
	try {
		sfpt = fopen(srcFileName, "rb");
		if (sfpt == NULL) {
			throw EXERR_NOTFOUND;
		}
		readBytes(sfpt, buf, 12);
		if (strncmp(buf, "GAMEDAT PAC", 11) != 0) {
			throw EXERR_BADFORMAT;
		}
		int dirLen;
		if (buf[11] == 'K') {
			dirLen = 16;
		} else if (buf[11] == '2') {
			dirLen = 32;
		} else {
			throw EXERR_BADFORMAT;
		}
		int entryCount = static_cast<int>(readUINT32(sfpt));
		if (entryCount <= 0) {
			throw EXERR_BADFORMAT;
		}
		std::vector<DirInfo> dirs;
		DirInfo entry;
		memset(&entry, 0, sizeof(entry));
		std::map<std::string, int> dmap;
		buf[dirLen] = '\0';
		int index;
		for (index = 0; index < entryCount; index++) {
			readBytes(sfpt, buf, dirLen);
			entry.name = std::string(buf);
			dirs.push_back(entry);
			dmap[entry.name] = index;
		}
		for (index = 0; index < entryCount; index++) {
			dirs[index].range.ofs = static_cast<int>(readUINT32(sfpt));
			dirs[index].range.len = static_cast<int>(readUINT32(sfpt));
		}
		long base = ftell(sfpt);
		if (base == -1) {
			throw EXERR_IOERROR;
		}
		dfpt = fopen(dstFileName, "wb");
		if (dfpt == NULL) {
			throw EXERR_NOTCREATE;
		}
		if (dirLen == 16) {
			writeBytes(dfpt, "GAMEDAT PACK", 12);
		} else {
			writeBytes(dfpt, "GAMEDAT PAC2", 12);
		}
		writeUINT32(dfpt, static_cast<uint32_t>(entryCount));
		for (index = 0; index < entryCount; index++) {
			strncpy(buf, dirs[index].name.c_str(), dirLen+1);
			writeBytes(dfpt, buf, dirLen);
		}
		std::vector<int> dirTable;
		dirTable.reserve(entryCount);
		for (index = 0; index < entryCount; index++) {
			int index2 = index;
			ConvertMap::const_iterator
				cit = cmap.find(dirs[index].name);
			if (cit != cmap.end()) {
				DirMap::const_iterator
					dit = dmap.find((*cit).second);
				if (dit != dmap.end()) {
					index2 = (*dit).second;
				}
			}
			dirTable.push_back(index2);
		}
		int ofs = 0;
		for (index = 0; index < entryCount; index++) {
			int index2 = dirTable[index];
			writeUINT32(dfpt, static_cast<uint32_t>(ofs));
			writeUINT32(dfpt, static_cast<uint32_t>(dirs[index2].range.len));
			ofs += dirs[index2].range.len;
		}
		for (index = 0; index < entryCount; index++) {
			int index2 = dirTable[index];
			if (fseek(sfpt, base+dirs[index2].range.ofs, SEEK_SET) != 0) {
				throw EXERR_IOERROR;
			}
			int left = dirs[index2].range.len;
			while (left > 0) {
				int copys = left;
				if (copys > sizeof(buf)) {
					copys = sizeof(buf);
				}
				readBytes(sfpt, buf, copys);
				writeBytes(dfpt, buf, copys);
				left -= copys;
			}
		}
		fclose(sfpt);
		sfpt = NULL;
		fclose(dfpt);
		dfpt = NULL;
	} catch (ExtractResultType /*error*/) {
		if (sfpt != NULL) {
			fclose(sfpt);
		}
		if (dfpt != NULL) {
			fclose(dfpt);
			unlink(dstFileName);
		}
		throw;
	}
}

///////////////////////////////////////////////////////////////////////////////
// prompt

int promptYesNo() {
	char buf[256];
	int result = -1;
	while (result < 0) {
		printf("(y/n) ");
		fflush(stdout);
		if (fgets(buf, sizeof(buf), stdin) == NULL) {
			return -1;
		}
		char ch;
		if (sscanf(buf, "%c", &ch) == 1) {
			if ((ch == 'y') || (ch == 'Y')) {
				result = 1;
			} else if ((ch == 'n') || (ch == 'N')) {
				result = 0;
			}
		}
	}
	return result;
}

///////////////////////////////////////////////////////////////////////////////
// main

int main(int argc, char* argv[]) {
	printf(
		"wpeBVGȂɂ񂱁x{CXւc[ PnyanVChg Ver. 1.00\n"
		"Programed by Manuke\n"
		"\n");
	std::string pnyanDir;
	bool help = false;
	for (int i = 1; i < argc; i++) {
		std::string arg = argv[i];
		if ((arg == "-h") || (arg == "--help")) {
			help = true;
		} else if ((arg == "-d") || (arg == "--dir")) {
			if (i+1 < argc) {
				pnyanDir = argv[++i];
			}
		}
	}
	if (help) {
		printf(
			"g:\n"
			"    PnyanVChg [option]\n"
			"\n"
			"IvV:\n"
			"    -h, --help\n"
			"        wv̕\ĩbZ[Wj\n"
			"    -d <dir>, --dir <dir>\n"
			"        wpeBVGȂɂ񂱁xfBNg̎w\n"
			"        iftHgFWXg莩擾j\n");
		return 0;
	}
	if (pnyanDir.empty()) {
		pnyanDir = getRegistryString(
			"Software\\Pajamas\\PNYAN",
			"InstallFolder",
			"");
	}
	if (pnyanDir.empty()) {
		printf("G[ : fBNg擾ł܂łB\n");
		return 1;
	}
	const char* front = pnyanDir.c_str();
	const char* back = CharPrev(front, front+pnyanDir.size());
	if (*back != '\\') {
		pnyanDir += '\\';
	}
	printf("fBNg : %s\n", pnyanDir.c_str());
	std::string targetFileName = pnyanDir+"archivesys.dat",
		tmpFileName = targetFileName+".tmp",
		orgFileName = targetFileName+".org";
	if (!existFile(targetFileName.c_str())) {
		printf("G[ : 'archivesys.dat'܂łB\n");
		return 1;
	}
	bool existOrg = existFile(orgFileName.c_str());
	if (!existOrg) {
		printf("VXe{CXւ܂H");
	} else {
		printf(
			"ɃVXe{CXւĂ܂B\n"
			"VXe{CXɖ߂܂H");
	}
	int result = promptYesNo();
	if (result != 1) {
		printf("𒆒f܂B\n");
		return 0;
	}
	if (!existOrg) {
		ConvertMap cmap;
		cmap["J770018.wav"] = "K710018.wav";
		cmap["J770019.wav"] = "K710019.wav";
		cmap["J770020.wav"] = "K710020.wav";
		cmap["J770021.wav"] = "K710021.wav";
		cmap["J770022.wav"] = "K710022.wav";
		cmap["J770023.wav"] = "K710023.wav";
		cmap["J770024.wav"] = "K710024.wav";
		cmap["J770026.wav"] = "K710026.wav";
		try {
			printf("t@C쐬...");
			fflush(stdout);
			convertPackFile(
				tmpFileName.c_str(),
				targetFileName.c_str(),
				cmap);
			printf("I\n");
		} catch (ExtractResultType error) {
			printf("\n");
			const char* message = NULL;
			switch (error) {
			case EXERR_NOTFOUND:
				message = "G[ : t@C܂łB\n";
				break;
			case EXERR_NOTCREATE:
				message = "G[ : t@C쐬ł܂łB\n";
				break;
			case EXERR_IOERROR:
				message = "G[ : I/OG[܂B\n";
				break;
			case EXERR_READERROR:
				message = "G[ : ǂݍ݃G[܂B\n";
				break;
			case EXERR_WRITEERROR:
				message = "G[ : ݃G[܂B\n";
				break;
			case EXERR_BADFORMAT:
				message = "G[ : tH[}bgsłB\n";
				break;
			}
			if (message != NULL) {
				printf(message);
			}
			return 1;
		}
		if (rename(targetFileName.c_str(), orgFileName.c_str()) != 0) {
			printf("G[ : l[Ɏs܂B\n");
			return 1;
		}
		if (rename(tmpFileName.c_str(), targetFileName.c_str()) != 0) {
			printf("G[ : l[Ɏs܂B\n");
			return 1;
		}
	} else {
		if (unlink(targetFileName.c_str()) != 0) {
			printf("G[ : 폜Ɏs܂B\n");
			return 1;
		}
		if (rename(orgFileName.c_str(), targetFileName.c_str()) != 0) {
			printf("G[ : l[Ɏs܂B\n");
			return 1;
		}
	}
	printf("܂B\n");
	return 0;
}
